#include "FolderView.h"
#include "FolderFactory.h"
#include "Util.h"
#include "Folder.h"

#define DEFAULT_SPACINGX 72
#define DEFAULT_SPACINGY 48

WNDPROC gListProc=0;

//=========================================================
// CFolder ///////////////////////////////////////////////
//=========================================================

CFolder::CFolder( CFolderFactory* owner ):
mBackColor(CLR_NONE),
mFont(NULL),
mFontColor(0),
mFlags(ffVisible),
mNext(NULL),
mOwner(owner)
{	
	mParent = 0;
	mRelOrigin.x = mRelOrigin.y = 0;
	
	mRightClickCommand[0] = 0;
	
	mFolder = NULL;
	mID = NULL;
	
	mView = NULL;
	mViewWnd = NULL;
}

CFolder::~CFolder()
{
	if(mNext)
		delete mNext;
	mNext = NULL;
	
	clear();

	if(mFolder)
	{
		mFolder->Release();
	}
	if(mID)
	{
		gMalloc->Free(mID);
	}
	
	if(mFont)
	{
		DeleteObject(mFont);
	}
	
	if(mWnd)
	{
		SetParent(mViewWnd, NULL);
		DestroyWindow(mWnd);
	}
}

DWORD CFolder::parseConfig(LPCSTR szLine)
{
	char token[MAX_LINE_LENGTH];
	LPCSTR nextToken = szLine;
	DWORD result = 0;
	bool hasCommand = true;
	static int cLastLeft = 0;
	static int cLastTop = 0;
	
	clear();
	
	GetToken(nextToken, NULL, &nextToken, false);
	
	if (GetToken(nextToken, token, &nextToken, false))
	{		
		mPosition.left = atoi(token);
		strlwr(token);
		if (strchr(token, 'r'))
			mPosition.left += cLastLeft;
	}
	if (GetToken(nextToken, token, &nextToken, false))
	{
		mPosition.top = atoi(token);
		strlwr(token);
		if (strchr(token, 'r'))
			mPosition.top += cLastTop;
	}
	cLastLeft = mPosition.left;
	cLastTop = mPosition.top;
	
	if (GetToken(nextToken, token, &nextToken, false))
	{
		mPosition.right = abs(atoi(token));
	}
	if (GetToken(nextToken, token, &nextToken, false))
	{
		mPosition.bottom = abs(atoi(token));
	}
	
	if (GetToken(nextToken, token, &nextToken, false))
	{
		LPSHELLFOLDER psfDesktop = NULL; // namespace root for parsing the path
		SHGetDesktopFolder(&psfDesktop);
		
		// IShellFolder::ParseDisplayName requires the path name in Unicode.
		OLECHAR olePath[MAX_PATH]; // wide-char version of path name
		MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, token, -1, olePath, MAX_PATH);
		
		// parse path for absolute PIDL, and connect to target folder
		
		if(mID)
			gMalloc->Free(mID);
		mID = NULL;
		
		if(!Resolve(NULL, psfDesktop, olePath, mID))
		{
			return 1;
		}		
	}
	
	if (GetToken(nextToken, token, &nextToken, false))
	{
		if (token[0] == '#')
		{
			strlwr(token);
			result = atoi(&token[1]);
			
			if(!strchr(token, 'h'))
				mFlags |= ffVisible;
			else mFlags &= ~ffVisible;
			
			if(strchr(token, 't') != 0)
				mFlags |= ffTopMost;
			else mFlags &= ~ffTopMost;
			
			if(strchr(token, 'b') != 0)
				mFlags |= ffHugBottom;
			else mFlags &= ~ffHugBottom;
			
			if(strchr(token, 'd') != 0)
				mFlags |= ffBorder;
			else mFlags &= ~ffBorder;
		}
	}
		
	if(GetToken(nextToken, token, &nextToken, false))
	{
		mRightClickCommand[0] = 0;
		do
		{
			if(mRightClickCommand[0])
				strcat(mRightClickCommand, " ");
			
			strcat(mRightClickCommand, token);
		} while(GetToken(nextToken, token, &nextToken, false));
	}
	else
		strcpy(mRightClickCommand, "!none");
	
	return result;
}

// any value that is <= 0 will not be changed
void CFolder::setPos(int x, int y, int cx, int cy)
{
	if(x > 0)
		mPosition.left = x;
	
	if(y > 0)
		mPosition.top = y;
	
	if(cx > 0)
		mPosition.right = cx;
	
	if(cy > 0)
		mPosition.bottom = cy;
	
	if(mNext)
		mNext->setPos(x,y,cx,cy);
}

void CFolder::setTopMost(int topmost)
{
	bool wasTopMost = (mFlags & ffTopMost) != 0;
	
	if (!topmost)
		mFlags ^= ffTopMost;
	else if(topmost > 0)
		mFlags |= ffTopMost;
	else
		mFlags &= ~ffTopMost;
	
	if (mViewWnd && ((mFlags & ffTopMost != 0) != wasTopMost))
	{
		// SetWindowLong(handle(), GWL_STYLE,
		// (GetWindowLong(handle(), GWL_STYLE) & ~WS_POPUP) | WS_CHILD);
		if ((mFlags & ffTopMost))
		{
			// SetParent(handle(), NULL);
			// SetWindowText(handle(), NULL);
			// SetWindowLong(handle(), GWL_STYLE,
			// (GetWindowLong(handle(), GWL_STYLE) & ~WS_CHILD) | WS_POPUP);
			SetWindowPos(mViewWnd, HWND_TOPMOST, 0, 0, 0, 0,
				SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
		}
		else
		{
			SetWindowPos(mViewWnd, HWND_NOTOPMOST, 0, 0, 0, 0,
				SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);
			// SetParent(handle(), desktopWnd);
			// SetWindowText(handle(), (fCaption ? fCaption : "CFolder"));
		}
	}
	
	if (mNext)
		mNext->setTopMost(topmost);
}

void CFolder::setBox(HWND wnd)
{
/*setVisible(1);

	error("wnd %p; flags %d", hWnd, mFlags);*/
	
	clear();
	
	mFlags |= ffVisible | ffDocked;
	mParent = wnd;
	
	createFolder();
	
	
	/*	int style = WS_CHILD;
	
	  
		if(wnd)
		mFlags |= ffDocked;
		else mFlags &= ~ffDocked;
		
		  SetWindowLong(hWnd, GWL_STYLE, style);
		  //	LowLevelUpdate(hWnd);
		  
			SetParent(hWnd, wnd);
			//	InvalidateRect(hWnd, NULL, TRUE);
	//	UpdateWindow(hWnd);*/
}

void CFolder::setVisible(int visible)
{
	bool wasVisible = (mFlags & ffVisible) != 0;
	
	if (!visible)
		mFlags ^= ffVisible;
	else if(visible > 0)
		mFlags |= ffVisible;
	else
		mFlags &= ~ffVisible;
	
	if (!mViewWnd && mFlags & ffVisible)
		createFolder();
	
	if ((mFlags & ffVisible != 0) != wasVisible)
		ShowWindow(mViewWnd, (mFlags & ffVisible ? SW_SHOW : SW_HIDE));
	
	if (mNext)
		mNext->setVisible(visible);
}

void CFolder::setBrowse (int browse)
{
	if (!browse)
		mFlags ^= ffBrowse;
	else if(browse > 0)
		mFlags |= ffBrowse;
	else
		mFlags &= ~ffBrowse;
	
	if (mNext)
		mNext->setBrowse(browse);
}

void CFolder::setFont(HFONT font)
{
	mFont = font;
	
	if(mNext)
		mNext->setFont(font);
}

void CFolder::setColors(COLORREF fontColor, COLORREF backColor)
{
	HWND folderView=  GetWindow(mViewWnd, GW_CHILD);

	if(backColor != CLR_NONE && mBackColor != backColor)
	ListView_SetBkColor(folderView,   backColor);
	if(mFontColor != fontColor)
	ListView_SetTextColor(folderView, fontColor);

	mFontColor = fontColor;
	mBackColor = backColor;
	

	if(mNext)
		mNext->setColors(fontColor, backColor);
}

void CFolder::setCommand(LPCSTR command)
{
	if(strcmp(command, "!none"))
		strcpy(mRightClickCommand, command);
	else
		mRightClickCommand[0] = 0;
	
	if(mNext)
		mNext->setCommand(command);
}

void CFolder::setPath (LPCSTR folder)
{		
	LPSHELLFOLDER psfDesktop = NULL; // namespace root for parsing the path
	SHGetDesktopFolder(&psfDesktop);
	
	// IShellFolder::ParseDisplayName requires the path name in Unicode.
	OLECHAR olePath[MAX_PATH]; // wide-char version of path name
	MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, folder, -1, olePath, MAX_PATH);
	
	// parse path for absolute PIDL, and connect to target folder
	
	LPITEMIDLIST oldidl = mID;
	mID = NULL;
				
	if(!Resolve(NULL, psfDesktop, olePath, mID))
	{
		mID = oldidl;
	}
	else
	{
		gMalloc->Free(oldidl);
	}		
	
	if(mNext)
		mNext->setPath(folder);
	
	refresh();
}

void CFolder::setBackground(LPCSTR background)
{ 
	if(background[0] && strcmp(background, "none"))
	{
		char path[1024];
		char buffer[1024];
		VarExpansion(buffer, background);
		
		if(strncmp(buffer, "http", 4))
		{
			strcpy(path,buffer);
			strcpy(buffer, "file:///");
			strcat(buffer, path);
		}
		
		for(char* pos = buffer; *pos; pos++)
			if(*pos == '\\')*pos = '/';
			
		LVBKIMAGE idata;
			
		idata.ulFlags = LVBKIF_SOURCE_URL | LVBKIF_STYLE_NORMAL | LVBKIF_STYLE_TILE;
		idata.xOffsetPercent = idata.yOffsetPercent = 0;
		idata.pszImage = buffer;
		
		
		ListView_SetBkImage(GetWindow(mViewWnd, GW_CHILD), &idata);
		
		mFlags |= ffBackground;
	}
	else
		mFlags &= ~ffBackground;
	
	if(mNext)
		mNext->setBackground(background);
}

void CFolder::browseUp()
{
	int size = 0, lastsize = 0;
	
	int count=0;
	
	LPSHITEMID mid = &mID->mkid;
	while(mid->cb != 0 && count < 1000)
	{
		count++;
		
		lastsize = size;
		size += mid->cb;
		mid = (LPSHITEMID)(((LPCH)mid) + mid->cb);
	}
	
	if(size == 0) // we are already at desktop, this stops flashing
	{
		refresh();
		return;
	}
	
	size = lastsize;
	
	LPSHITEMID newID = (LPSHITEMID)gMalloc->Alloc(size + sizeof(LPSHITEMID));
	
	if(size)
		memcpy(newID, mID, size);
	LPSHITEMID terminator = LPSHITEMID(((char*)newID) + size);
	
	terminator->abID[0] = 0;
	terminator->cb = 0;
	
	gMalloc->Free(mID);
	mID = (LPITEMIDLIST)newID;
	refresh();
}

void CFolder::browseDown(LPCSTR subfolder)
{
	error("Browse down support not written yet");
}

void CFolder::refresh()
{
	if(mView)
	{
		SetParent(mViewWnd, NULL);

		mView->DestroyViewWindow();
		mViewWnd= NULL;
		
		mView->Release();
		mView = NULL;
	}
	
	if(mFolder)
	{
		mFolder->Release();
		mFolder = NULL;
	}
	
	
	HRESULT hr; 
	
	if(!mID)
		return;
	
	LPSHELLFOLDER psfDesktop = NULL; // namespace root for parsing the path
	SHGetDesktopFolder(&psfDesktop);
	
	if(mID->mkid.cb == 0)
	{
		mFolder = psfDesktop;
	}
	else
	{
		hr = psfDesktop->BindToObject(mID, NULL, IID_IShellFolder, 
			(void**)&mFolder);
		psfDesktop->Release(); // no longer required
		
		if(FAILED(hr))
		{
			return;
		}
	}
	
	mFolder->CreateViewObject(mWnd, IID_IShellView, (void**)&mView);
	
	if(!mView)
	{
		return;
	}
	
	FOLDERSETTINGS fs;
	
	fs.ViewMode = FVM_ICON;
	fs.fFlags = FWF_NOCLIENTEDGE ;

//	if(mBackColor == CLR_NONE && (mFlags & ffBackground) == 0)
		fs.fFlags |= FWF_DESKTOP;
	
	RECT window = { 0, 0, mPosition.right, mPosition.bottom };
//	OffsetRect(&window, mPosition.left, mPosition.top);
	
	mView->CreateViewWindow(NULL, &fs, (IShellBrowser*)gFolderFactory, &window, &mViewWnd);
	
	gFolderFactory->mLastView = mView;

	mView->UIActivate(SVUIA_ACTIVATE_NOFOCUS);

	if(!mViewWnd)
		error("Failed to create view window");

	SetParent(mViewWnd, mWnd);

	HWND folderView = GetWindow(mViewWnd, GW_CHILD);
	
	gListProc = (WNDPROC)SetWindowLong(folderView, GWL_WNDPROC, (LONG)ViewWndProc);
	
	SetWindowLong(folderView, GWL_USERDATA, (LONG)this);
	LowLevelUpdate(folderView);

	ListView_SetTextBkColor(folderView,   CLR_NONE);

	if(mBackColor != CLR_NONE)
	ListView_SetBkColor(folderView,   mBackColor);
	ListView_SetTextColor(folderView, mFontColor);

	if(mFont)
	SetFont(folderView, mFont);

	InvalidateRect(folderView, NULL, TRUE);
	UpdateWindow(folderView);

	ShowWindow(mViewWnd, SW_SHOW);
}

void CFolder::clear()
{	
	if(mID)
		gMalloc->Free(mID);
	mID = NULL;
	
	if(mWnd)
	{
		SetParent(mViewWnd, NULL);
		DestroyWindow(mWnd);
		mWnd = NULL;
	}

	if(mView)
	{
		SetParent(mViewWnd, NULL);
		
		mView->DestroyViewWindow();
		mViewWnd= NULL;
		
		mView->Release();
		mView = NULL;
	}
	
	if(mFolder)
	{
		mFolder->Release();
		mFolder = NULL;
	}
}


void CFolder::createFolder()
{	
	mRelOrigin = convertCoords(mPosition.left, mPosition.top);

	if(mWnd)
	{
		SetParent(mViewWnd, NULL);
		DestroyWindow(mWnd);
	}

	mWnd = CreateWindow(FOLDER_CLASSNAME,
		"Folder",
		WS_VISIBLE | ((mFlags & ffBorder) == 0 ? WS_POPUP : WS_BORDER),
		mPosition.left, mPosition.top,
		mPosition.right, mPosition.bottom,
		mParent ? mParent : gFolderFactory->handle(), NULL,
		Window::hInstance,
		0);	

	SetWindowLong(mWnd, GWL_USERDATA, magicDWord);
	SetWindowLong(mWnd, 0, LONG(this));

	refresh();	

	if(mFlags & ffTopMost)
	{
		SetWindowPos(mWnd, HWND_TOPMOST, 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE);
	}
	else if(mFlags & ffHugBottom)
	{
		SetWindowPos(mWnd, gPlacementWnd, 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE);
	}
	else
	{
		SetWindowPos(mWnd, HWND_TOP, 0,0,0,0, SWP_NOMOVE | SWP_NOSIZE);
	}
}


POINT CFolder::convertCoords(const int left, const int top) const
{
	POINT relPt;
	
	relPt.x = CONVERT_COORDINATE_X(left);
	relPt.y = CONVERT_COORDINATE_Y(top);
	
	if (!MonitorFromPoint(relPt, MONITOR_DEFAULTTONULL))
	{
		RECT rcMon;
		MONITORINFO mi;
		mi.cbSize = sizeof(mi);
		
		GetMonitorInfo(MonitorFromPoint(relPt, MONITOR_DEFAULTTONEAREST), &mi);
		rcMon = mi.rcMonitor;
		if (relPt.x < rcMon.left)
			relPt.x = rcMon.left;
		else if (relPt.x > rcMon.right)
			relPt.x = rcMon.right;
		if (relPt.y < rcMon.top)
			relPt.y = rcMon.top;
		else if (relPt.y > rcMon.bottom)
			relPt.y = rcMon.bottom;
	}
	return relPt;
}


//---------------------------------------------------------
// Message handling
//---------------------------------------------------------

/*void CFolder::windowProc(Message& message)
{
	BEGIN_MESSAGEPROC
	case WM_GETDLGCODE: 
	{
		message.lResult = DLGC_WANTALLKEYS; 
	} break;
	MESSAGE(onCreate, WM_CREATE)
		MESSAGE(onDestroy, WM_DESTROY)
		MESSAGE(onEndSession, WM_ENDSESSION)
		MESSAGE(onEndSession, WM_QUERYENDSESSION)
		MESSAGE(onSysCommand, WM_SYSCOMMAND)
		//MESSAGE(onMouseActivate, WM_MOUSEACTIVATE)
		MESSAGE(onEraseBkgnd, WM_ERASEBKGND)
		MESSAGE(onPosChanging, WM_WINDOWPOSCHANGING)
		MESSAGE(onMove, WM_MOVE)
		MESSAGE(onActivate, WM_ACTIVATE)
		END_MESSAGEPROC
}

void CFolder::onActivate(Message& msg)
{
	if(mView && (msg.wParamLo == WA_ACTIVE || msg.wParamLo == WA_CLICKACTIVE))
		gFolderFactory->mLastView = mView;
}

void CFolder::onMove(Message& message)
{
	InvalidateRect(hWnd, NULL, TRUE);
}

void CFolder::onCreate(Message& message)
{
	SetWindowLong(hWnd, GWL_USERDATA, magicDWord);
}

void CFolder::onDestroy(Message& message)
{
	if(mParent)
	{
		SetParent(hWnd, NULL);
		mParent = NULL;
	}
	
	clear();
	destroyWindow();
	
	mFlags &= ~(ffDocked | ffVisible);
}


void CFolder::onEndSession(Message& message)
{
	message.lResult = SendMessage(GetLitestepWnd(), message.uMsg, message.wParam, message.lParam);
}


void CFolder::onSysCommand(Message& message)
{
	if (message.wParam == SC_CLOSE)
		PostMessage(GetLitestepWnd(), WM_KEYDOWN, LM_SHUTDOWN, 0);
	else
		message.lResult = DefWindowProc(hWnd, message.uMsg, message.wParam, message.lParam);
}

void CFolder::onMouseActivate(Message& message)
{
	SetFocus(mViewWnd);
	PostMessage(hWnd, WM_LBUTTONDOWN, 0, 0);
	message.lResult = MA_ACTIVATE;
}

void CFolder::onPosChanging(Message& message)
{
	if(!(mFlags & ffDocked) && mFlags & ffHugBottom)
	{
		LPWINDOWPOS wp = (LPWINDOWPOS)message.lParam;
		wp->hwndInsertAfter = gPlacementWnd;
	}
	
	message.lResult = 0;
}

void CFolder::onEraseBkgnd(Message& message)
{	
	//if(mFlags & ffDocked)
	//	return;
	
	HDC hDC = (HDC)message.wParam;
	
	if((mFlags & ffBackground) == 0)
	{
		if(mBackColor == CLR_NONE)
		{ 	
			if(mFlags & ffDocked)
			{
				return;
					PaintDesktopEx(hDC, 0, 0, 
				mPosition.right, mPosition.bottom, mPosition.left,
				mPosition.top, FALSE);
			}
			else
				PaintDesktop(hDC);
		}
		else
		{
			HBRUSH back = CreateSolidBrush(mBackColor);
			
			RECT dest = {0,0,mPosition.right,mPosition.bottom};
			FillRect(hDC, &dest, back);
			
			DeleteObject(back);
		}
	}
}

bool CFolder::onMouseRightClick()
{
	if(mRightClickCommand[0] && strcmp(mRightClickCommand, "!none"))
	{
		LSExecute(hWnd, mRightClickCommand, SW_SHOW);
		return true;
	}
	return false;		
}*/

LRESULT CALLBACK CFolder::ViewWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	CFolder * owner = (CFolder*)GetWindowLong(hWnd, GWL_USERDATA);
	
	switch(uMsg)
	{
	case WM_ERASEBKGND:
		{
			HDC hDC = (HDC)wParam;
			
			if(owner)
			{	//if(mFlags & ffDocked)
				//	return;
				
				if(owner->mFlags & ffBackground)
					return CallWindowProc(gListProc, hWnd, uMsg, wParam, lParam);
				else 
				{
					if(owner->mBackColor == CLR_NONE)
					{ 	
						if(!(owner->mFlags & ffDocked))
						{
							PaintDesktop(hDC);
						}
					}
					else
					{
						HBRUSH back = CreateSolidBrush(owner->mBackColor);
						
						RECT dest = {0,0,owner->mPosition.right,owner->mPosition.bottom};
						FillRect(hDC, &dest, back);
						
						DeleteObject(back);
					}
				}
				
			} 
			
			return 1;
		} break;
	case WM_DESTROY:
		{
			SetWindowLong(hWnd, GWL_WNDPROC, (LONG)gListProc);
		} break;
	}
	
	return CallWindowProc(gListProc, hWnd, uMsg, wParam, lParam);
}

