#ifndef FOLDERVIEW_FOLDER_H
#define FOLDERVIEW_FOLDER_H

enum 
{
	ffBrowse   = 0x01,
	ffScroll   = 0x02,
	ffVisible  = 0x04,
	ffHugBottom= 0x08,
	ffTopMost  = 0x10,
	ffBorder   = 0x20,
	ffDocked   = 0x40,
	ffBackground = 0x80
};

class CFolder 
{
	friend LRESULT CALLBACK WndProc(HWND wnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
	friend class CFolderFactory;
	
public:
	CFolder(CFolderFactory* owner);
	~CFolder();
	
	inline void Start(DWORD &group, LPCSTR config)
	{
		group = parseConfig(config);

		if(mFlags & ffVisible)
			createFolder();
		
		refresh();
	}

	// any value that is <= 0 will not be changed
	void setPos(int x, int y, int cx, int cy);
	
	void setTopMost(int topmost);
	void setVisible(int visible);
	void setBrowse (int browse);

	void setFont(HFONT font);
	void setColors(COLORREF fontColor, COLORREF backColor);
	
	void setCommand(LPCSTR command);
	void setPath (LPCSTR folder);
	void setBackground(LPCSTR background);

	void browseUp();
	void browseDown(LPCSTR subfolder);

	void refresh();

	void setBox(HWND wnd);
	
protected:

	void clear();

	DWORD parseConfig(LPCSTR szLine);
	void createFolder();
	
	POINT convertCoords(const int left, const int top) const;
		
	CFolderFactory* mOwner;
	
	char mRightClickCommand[512];

	POINT mRelOrigin;
	RECT mPosition; // left, top, width, height
	
	byte mFlags;
	HWND mParent;

	HFONT mFont;
	COLORREF mFontColor, mBackColor;

	LPSHELLFOLDER mFolder;
	LPITEMIDLIST  mID;

	LPSHELLVIEW   mView;
	HWND          mViewWnd;

	HWND  mWnd;

	CFolder* mNext;

	// message handling
/*	virtual void windowProc(Message& message);

	void onCreate(Message& message);
	void onDestroy(Message& message);
	void onEndSession(Message& message);
	void onSysCommand(Message& message);

	void onActivate(Message& msg);
	void onMouseActivate(Message& message);
	bool onMouseRightClick();

	void onEraseBkgnd(Message& message);
	void onPosChanging(Message& message);

	void onMove(Message& message);*/
	
	static LRESULT CALLBACK ViewWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
};

#endif