#include "FolderView.h"
#include "Util.h"
#include "FolderFactory.h"
#include "Folder.h"

CFolderFactory* gFolderFactory=0;

enum MESSAGES
{
	UM_SETTOPMOST = WM_USER,
	UM_UPFOLDER
};

//=========================================================
// folderFactory ////////////////////////////////////////
//=========================================================

CFolderFactory::CFolderFactory(HWND parentWnd):Window(APP_NAME)
{
	createWindow(WS_EX_TOOLWINDOW, "folderFactory", WS_POPUP,
		0, 0, 0, 0, parentWnd);
	SetWindowLong(hWnd, GWL_USERDATA, magicDWord);
}


CFolderFactory::~CFolderFactory()
{
//	DeleteObject(mFont);

	destroyFolders();
	destroyWindow();
}

void CFolderFactory::setDefaults()
{
	mBrowse = false;
	mScrollBars = false;

	mFont = NULL;
	
	mFontColor = 0xFFFFFFFF;
	mBackColor = CLR_NONE;
	mBackgroundImage[0] = 0;
}

// input variables into the tokeniser
const char* ids[] = {
	"scroll",
	"font",
	"fontColor",
	"backColor",
	"browse",
	"back",
	""
};

enum TokenID
{
	tokScroll,
	tokFont,
	tokFontColor,
	tokBackColor,
	tokBrowse,
	tokBackground
};

void CFolderFactory::parseSettings(LPCSTR nextToken, LPSTR token)
{
	GetToken(nextToken, token, &nextToken, true);
	char* args;
	
	int tok = -1;
	while((tok=ConvertToken(token, ids, &args)) != -1)
	{
		switch(tok)
		{
		case tokScroll:
			{
				if(strcmp(args, "on"))
					mScrollBars = false;
				else
					mScrollBars = true;
			} break;
			
		case tokFont:
			{
				int pt = 16;
				int weight = FW_NORMAL;
				int italics = 0, underline = 0, strikeout = 0;
				
				char *pos = strchr(args, '_');
				if(pos)
				{
					*pos++ = 0;
					
					strlwr(pos);
					
					char* stopchar;
					pt = strtol(pos, &stopchar, 10);
					
					if(stopchar && *stopchar)
					{
						while(*stopchar != 0 && stopchar - pos < 1000)
						{
							if(*stopchar == 'b')
								weight = FW_BOLD;
							if(*stopchar == 'i')
								italics = TRUE;
							if(*stopchar == 'u')
								underline = TRUE;
							if(*stopchar == 's')
								strikeout = TRUE;
							
							stopchar++;
						}
					}
				}
				
				
				mFont = CreateFont(
					pt,
					0,0,0,
					weight,
					italics,underline, strikeout,
					ANSI_CHARSET,
					OUT_TT_PRECIS,
					CLIP_DEFAULT_PRECIS,
					PROOF_QUALITY,
					DEFAULT_PITCH | FF_DONTCARE,
					args
					);
			} break;
			
		case tokFontColor:
			{
				if(strcmp(args, "none"))
				{
					mFontColor = strtol(args, NULL, 16);
					//	mFontColor = RGB(GetBValue(fFontColor), GetGValue(fFontColor), GetRValue(fFontColor));
				}
				else
					mFontColor = CLR_NONE;
			} break;
			
		case tokBackColor:
			{
				if(strcmp(args, "none"))
				{
					mBackColor = strtol(args, NULL, 16);
				}
				else
					mBackColor = CLR_NONE;
			} break;
			
		case tokBrowse:
			{
				if(strcmp(args, "on"))
					mBrowse = false;
				else
					mBrowse = true;
			} break;
			
		case tokBackground:
			{
				if(strcmp(args, "none"))
					strcpy(mBackgroundImage, args);
				else
					mBackgroundImage[0] = 0;
			} break;
		}
		
		if(!GetToken(nextToken, token, &nextToken, true))
			break;
	}				
}
	
void CFolderFactory::createFolders()
{
	char configLine[MAX_LINE_LENGTH];
	FILE* step;
	DWORD groupNumber;
	CFolderMap::iterator smIter;
	CFolder* newFolder;
		
	setDefaults();
	
	step = LCOpen(NULL);
	if (step)
	{
		while (LCReadNextConfig(step, "*folder", configLine, MAX_LINE_LENGTH))
		{
			char token[MAX_LINE_LENGTH];
			LPCSTR nextToken = configLine;

			GetToken(configLine, NULL, &nextToken, false);
			if (GetToken(nextToken, token, &nextToken, false) && strcmp(token, "****") == 0)
			{
				parseSettings(nextToken, token);				
			}
			else
			{ 
				newFolder = new CFolder(this);
				if(!newFolder)
					continue;

				newFolder->setFont(mFont);

				newFolder->setBrowse(mBrowse? 1 : -1);

				newFolder->Start(groupNumber, configLine);
				newFolder->refresh();
	
				newFolder->setColors(mFontColor, mBackColor);
				newFolder->setBackground(mBackgroundImage);
				
				smIter = mFolderGroups.find(groupNumber);
				if (smIter == mFolderGroups.end())
					mFolderGroups.insert(CFolderMap::value_type(groupNumber, newFolder));
				else
				{
					if(newFolder->mNext)
						error("LINKED LIST ERROR");
					
					newFolder->mNext = smIter->second;
					smIter->second = newFolder;
				}
			}
		}
		LCClose(step);
	}
}


void CFolderFactory::destroyFolders()
{
	CFolderMap::iterator smIter;
	
	int count = 0;
	for (smIter = mFolderGroups.begin(); smIter != mFolderGroups.end(); smIter++)
	{
		count ++;

		CFolder* folder = smIter->second;

		if(folder)
			delete folder;
	}
	mFolderGroups.clear();
}


void CFolderFactory::setVisible(LPCSTR groups, int state)
{
	char group[MAX_LINE_LENGTH];
	LPCSTR nextGroup = groups;
	CFolderMap::iterator smIter;
	
	if (CharInStr('*', groups))
	{
		for (smIter = mFolderGroups.begin(); smIter != mFolderGroups.end(); smIter++)
		{
			if (smIter->second)
				smIter->second->setVisible(state);
		}
	}
	else
		while (GetToken(nextGroup, group, &nextGroup, false))
		{
			smIter = mFolderGroups.find(atoi(group));
			if (smIter != mFolderGroups.end())
				if (smIter->second)
					smIter->second->setVisible(state);
		}
}


//---------------------------------------------------------
// Bang commands
//---------------------------------------------------------

void CFolderFactory::bangBoxHook(HWND sender, LPCSTR args)
{
	char token[MAX_LINE_LENGTH];
	LPCTSTR nextToken[MAX_LINE_LENGTH];

	GetToken(args,token,nextToken,false);

	CFolderMap::iterator it = gFolderFactory->mFolderGroups.find(atoi(token));

	if(it == gFolderFactory->mFolderGroups.end())
		return;

	CFolder* folder = it->second;

	char *handle = strrchr(args,' ');
	
	if (folder  && handle)
	{
		HWND boxWnd = (HWND)atoi(handle+1);
		if (boxWnd) 
		{
			folder->setBox(boxWnd);
		}
		folder->refresh();
	}
}

void CFolderFactory::bangGroupToggle(HWND sender, LPCSTR args)
{
	gFolderFactory->setVisible(args, 0);
}


void CFolderFactory::bangGroupShow(HWND sender, LPCSTR args)
{
	gFolderFactory->setVisible(args, 1);
}


void CFolderFactory::bangGroupHide(HWND sender, LPCSTR args)
{
	gFolderFactory->setVisible(args, -1);
}


void CFolderFactory::bangGroupOnTopToggle(HWND sender, LPCSTR args)
{
	int state = 0;

	SendMessage(gFolderFactory->handle(), UM_SETTOPMOST, WPARAM(args), LPARAM(state));
}


void CFolderFactory::bangGroupOnTop(HWND sender, LPCSTR args)
{
	int state = 1;
	
	SendMessage(gFolderFactory->handle(), UM_SETTOPMOST, WPARAM(args), LPARAM(state));
}


void CFolderFactory::bangGroupOnBottom(HWND sender, LPCSTR args)
{
	int state = -1;

	SendMessage(gFolderFactory->handle(), UM_SETTOPMOST, WPARAM(args), LPARAM(state));
}

void CFolderFactory::bangGroupUp(HWND sender, LPCSTR args)
{
	int state = 0;

	SendMessage(gFolderFactory->handle(), UM_UPFOLDER, WPARAM(args), LPARAM(state));
}

//---------------------------------------------------------
// Message handling
//---------------------------------------------------------

void CFolderFactory::windowProc(Message& message)
{
	BEGIN_MESSAGEPROC
		MESSAGE(onCreate, WM_CREATE)
		MESSAGE(onDestroy, WM_DESTROY)
		MESSAGE(onGetRevId, LM_GETREVID)
	//	MESSAGE(onRefresh, LM_REFRESH)
		MESSAGE(onSetTopmost, UM_SETTOPMOST)
		MESSAGE(onGroupUp, UM_UPFOLDER)
		MESSAGE(onGetIsShellBrowser, WM_USER + 7)
	END_MESSAGEPROC
}


void CFolderFactory::onCreate(Message& message)
{
	int lsMessages[] = {LM_GETREVID, 0};
	
	SendMessage(hParent, LM_REGISTERMESSAGE, WPARAM(hWnd), LPARAM(lsMessages));
	
	AddBangCommand("!foldergrouptoggle", bangGroupToggle);
	AddBangCommand("!foldergroupshow", bangGroupShow);
	AddBangCommand("!foldergrouphide", bangGroupHide);
	AddBangCommand("!foldergroupontoptoggle", bangGroupOnTopToggle);
	AddBangCommand("!foldergroupontop", bangGroupOnTop);
	AddBangCommand("!foldergrouponbottom", bangGroupOnBottom);
	AddBangCommand("!folderup", bangGroupUp);

	AddBangCommand("!FolderLSBoxHook", bangBoxHook);
}


void CFolderFactory::onDestroy(Message& message)
{
	CFolderMap::iterator smIter;
	int lsMessages[] = {LM_GETREVID, 0};
	
	destroyFolders();
		
	SendMessage(hParent, LM_UNREGISTERMESSAGE, WPARAM(hWnd), LPARAM(lsMessages));
	
	RemoveBangCommand("!foldergrouptoggle");
	RemoveBangCommand("!foldergroupshow");
	RemoveBangCommand("!foldergrouphide");
	RemoveBangCommand("!foldergroupontoptoggle");
	RemoveBangCommand("!foldergroupontop");
	RemoveBangCommand("!foldergrouponbottom");
	RemoveBangCommand("!folderup");
	
	RemoveBangCommand("!FolderLSBoxHook");
}


void CFolderFactory::onGetIsShellBrowser(Message& message)
{
	message.lResult = (LRESULT)(IShellBrowser*)this;
}

void CFolderFactory::onGetRevId(Message& message)
{
	char* buf = (char*)(message.lParam);
	
	switch (message.wParam)
	{
	case 0:
	case 1:
		strcpy(buf, "FolderView.dll: 1.1 (Smoogle)");
		break;
	default:
		buf[0] = 0;
	}
	message.lResult = strlen(buf);
}

void CFolderFactory::onRefresh(Message& message)
{
//	destroyFolders();
//	createFolders();
}

void CFolderFactory::onGroupUp(Message& message)
{
	LPCSTR groups = LPCSTR(message.wParam);
	int topMost = int(message.lParam);
	char group[MAX_LINE_LENGTH];
	LPCSTR nextGroup = groups;
	CFolderMap::iterator smIter;
	
	if (CharInStr('*', groups))
	{
		for (smIter = mFolderGroups.begin(); smIter != mFolderGroups.end(); smIter++)
			if (smIter->second)
				smIter->second->browseUp();
	}
	else
		while (GetToken(nextGroup, group, &nextGroup, false))
		{
			smIter = mFolderGroups.find(atoi(group));
			if (smIter != mFolderGroups.end())
				if (smIter->second)
				smIter->second->browseUp();
		}
}

void CFolderFactory::onSetTopmost(Message& message)
{
	LPCSTR groups = LPCSTR(message.wParam);
	int topMost = int(message.lParam);
	char group[MAX_LINE_LENGTH];
	LPCSTR nextGroup = groups;
	CFolderMap::iterator smIter;
	
	if (CharInStr('*', groups))
	{
		for (smIter = mFolderGroups.begin(); smIter != mFolderGroups.end(); smIter++)
			if (smIter->second)
				smIter->second->setTopMost(topMost);
	}
	else
		while (GetToken(nextGroup, group, &nextGroup, false))
		{
			smIter = mFolderGroups.find(atoi(group));
			if (smIter != mFolderGroups.end())
				if (smIter->second)
					smIter->second->setTopMost(topMost);
		}
}