#ifndef FOLDERVIEW_FOLDER_FACTORY_H
#define FOLDERVIEW_FOLDER_FACTORY_H

class CFolder;

typedef map<UINT, CFolder*> CFolderMap;

class IShellBrowserImpl : public IShellBrowser, public ICommDlgBrowser
{
	DWORD m_dwRef;
public:
	IShellBrowserImpl():m_dwRef(0){}

	STDMETHOD(QueryInterface)(REFIID iid, void **ppvObject)
	{
		if(ppvObject == NULL)
			return E_POINTER;

		*ppvObject = NULL;

		if(iid == IID_IUnknown)
			*ppvObject = (IUnknown*)(IShellBrowser*) this;
		else if(iid == IID_IOleWindow)
			*ppvObject = (IOleWindow*) this;			
		else if(iid == IID_IShellBrowser)
			*ppvObject = (IShellBrowser*) this;
		else if(iid == IID_ICommDlgBrowser)
			*ppvObject = (ICommDlgBrowser*) this;
		else
			return E_NOINTERFACE;
		((IUnknown*)(*ppvObject))->AddRef();
		return S_OK;
	}
	
	 STDMETHOD(SendControlMsg)(THIS_ UINT id, UINT uMsg, WPARAM wParam,
		 LPARAM lParam, LRESULT * pret) { return E_NOTIMPL; }
	 STDMETHOD(GetControlWindow)(THIS_ UINT id, HWND * lphwnd) { return E_NOTIMPL; }

	STDMETHOD_(ULONG, AddRef)() { return ++m_dwRef; }
	STDMETHOD_(ULONG, Release)(){ return --m_dwRef; }  //not heap based	
    // *** IOleWindow methods ***
    STDMETHOD(ContextSensitiveHelp)(BOOL fEnterMode){return E_NOTIMPL;}	
	// *** ICommDlgBrowser methods ***
    STDMETHOD(OnDefaultCommand) (THIS_ struct IShellView * ppshv)
	{	//handle double click and ENTER key if needed
		return E_NOTIMPL; 
	}
    STDMETHOD(OnStateChange) (THIS_ struct IShellView * ppshv,ULONG uChange)
	{	//handle selection, rename, focus if needed
		return E_NOTIMPL; 
	}
    STDMETHOD(IncludeObject) (THIS_ struct IShellView * ppshv,LPCITEMIDLIST pidl)
	{	//filter files if needed
		return S_OK;
	}
    // *** IShellBrowser methods *** (same as IOleInPlaceFrame)
    STDMETHOD(InsertMenusSB) (HMENU hmenuShared, LPOLEMENUGROUPWIDTHS lpMenuWidths) {return E_NOTIMPL;}
    STDMETHOD(SetMenuSB) (HMENU hmenuShared, HOLEMENU holemenuReserved,HWND hwndActiveObject){return E_NOTIMPL;}
    STDMETHOD(RemoveMenusSB) (HMENU hmenuShared){return E_NOTIMPL;}
    STDMETHOD(SetStatusTextSB) (LPCOLESTR lpszStatusText){return E_NOTIMPL;}
    STDMETHOD(EnableModelessSB) (BOOL fEnable){return E_NOTIMPL;}
	STDMETHOD(BrowseObject)(LPCITEMIDLIST pidl, UINT wFlags){return E_NOTIMPL; }
	STDMETHOD(GetViewStateStream)(DWORD grfMode,LPSTREAM  *ppStrm){return E_NOTIMPL;}	
	STDMETHOD(OnViewWindowActive)(struct IShellView *ppshv){return E_NOTIMPL;}
	STDMETHOD(SetToolbarItems)(LPTBBUTTON lpButtons, UINT nButtons,UINT uFlags){return E_NOTIMPL;}
	STDMETHOD(TranslateAcceleratorSB) (LPMSG lpmsg, WORD wID) { return S_OK; }
};

class CFolderFactory : public Window, public IShellBrowserImpl
{
protected:
	CFolderMap mFolderGroups;

	bool mBrowse;
	bool mScrollBars;
	HFONT mFont;

	COLORREF mFontColor, mBackColor;
	char mBackgroundImage[512];

public:

	LPSHELLVIEW mLastView;

	void setDefaults();
	void parseSettings(LPCSTR nextToken, LPSTR token);

	CFolderFactory(HWND parentWnd);
	~CFolderFactory();

	void createFolders();
	void destroyFolders();
protected:

	void setVisible(LPCSTR groups, int state);
	void setTopMost(LPCSTR groups, int topMost);

	static void bangGroupToggle(HWND sender, LPCSTR args);
	static void bangGroupShow(HWND sender, LPCSTR args);
	static void bangGroupHide(HWND sender, LPCSTR args);
	static void bangGroupOnTop(HWND sender, LPCSTR args);
	static void bangGroupOnBottom(HWND sender, LPCSTR args);
	static void bangGroupOnTopToggle(HWND sender, LPCSTR args);
	static void bangGroupUp(HWND sender, LPCSTR args);
	static void bangGroupGoto(HWND sender, LPCSTR args);

	static void bangBoxHook(HWND sender, LPCSTR args);

	virtual void windowProc(Message& message);
	void onCreate(Message& message);
	void onDestroy(Message& message);

	void onGetRevId(Message& message);
	void onRefresh(Message& message);
	
	void onSetTopmost(Message& message);
	void onGroupUp(Message& message);
	void onGroupGoto(Message& Message);

	void onGetIsShellBrowser(Message& Message);

	STDMETHOD(GetWindow)(HWND * lphwnd)
	{ 
		*lphwnd = hWnd; 
		return S_OK; 
	}
	//IShellBrowser
	STDMETHOD(QueryActiveShellView)(struct IShellView ** ppshv)
	{
		if(!mLastView)
			return E_FAIL;

		mLastView->AddRef();
		*ppshv = mLastView;
		return S_OK; 
	}

	STDMETHOD(OnViewWindowActive)(struct IShellView *ppshv)
	{
		mLastView = ppshv;
		return S_OK;
	}

	STDMETHOD(BrowseObject)(LPCITEMIDLIST pidl, UINT wFlags)
	{
		return NOERROR; 
	}
};

extern CFolderFactory* gFolderFactory;

#endif