

//---------------------------------------------------------
// Global defs and vars
//---------------------------------------------------------


#include "FolderView.h"
#include "FolderFactory.h"
#include "Folder.h"
#include "Util.h"

HWND gDesktopWnd = 0;
HWND gPlacementWnd=0;

//---------------------------------------------------------
// Module startup and shutdown
//---------------------------------------------------------

LRESULT CALLBACK WndProc(HWND wnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

int initModuleEx(HWND parentWnd, HINSTANCE hInst, LPCSTR)
{
	Window::init(hInst);
	
	InitCommonControls();

	OleInitialize(NULL);	
	CoInitialize(NULL); // initialize COM

	if(gMalloc)
		gMalloc->Release();
	gMalloc = NULL;
	
	HRESULT hr = SHGetMalloc(&gMalloc); 
	if(FAILED(hr))
	{
		return 1;
	}

	gDesktopWnd = FindWindow("DesktopBackgroundClass", NULL);
	if (!gDesktopWnd)
		gDesktopWnd = GetDesktopWindow();
	
	gPlacementWnd = GetNextWindow(gDesktopWnd,GW_HWNDPREV);

	WNDCLASS wc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = sizeof(Folder*);
	wc.hbrBackground = NULL;
	wc.hCursor  = NULL;
	wc.hIcon = NULL;
	wc.hInstance = hInst;
	wc.lpfnWndProc = WndProc;
	wc.lpszClassName = FOLDER_CLASSNAME;
	wc.lpszMenuName = NULL;
	wc.style = CS_DBLCLKS;

	RegisterClass(&wc);

		if(gFolderFactory)
			delete gFolderFactory;
		
		gFolderFactory = new CFolderFactory(parentWnd);
		if(gFolderFactory)
		{
			gFolderFactory->createFolders();
			return 0;
		}
		
	return 1;
}

void quitModule(HINSTANCE inst)
{	
	if(gFolderFactory)
	delete gFolderFactory;

	gFolderFactory = NULL;

	UnregisterClass(FOLDER_CLASSNAME, inst);
	
	if(gMalloc)
		gMalloc->Release();

	gMalloc = NULL;
	
	CoUninitialize(); // shut down COM
	OleUninitialize();
}


LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	CFolder * owner = (CFolder*)GetWindowLong(hWnd, 0);

	switch(uMsg)
	{
	case WM_WINDOWPOSCHANGING:
		{
			if(owner && owner->mFlags & ffHugBottom)
				((WINDOWPOS*)lParam)->hwndInsertAfter = gPlacementWnd;
			return 1;
		} break;
	case WM_MOUSEACTIVATE:
		{
			if(owner)
				SetFocus(owner->mViewWnd);

			PostMessage(hWnd, WM_LBUTTONDOWN, 0, 0);
			return MA_ACTIVATE;
		} break;
	case WM_ACTIVATE:
		{
			if(owner && owner->mView && (LOWORD(wParam) == WA_ACTIVE || LOWORD(wParam) == WA_CLICKACTIVE))
				gFolderFactory->mLastView = owner->mView;
		} break;
}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}