#include "FolderView.h"
#include "Util.h"

LPMALLOC gMalloc=0; // memory manager, for freeing up PIDLs

// tokeniser:
int ConvertToken(char* token, const char* idlist[], char** args)
{
	if(token[0] == 0)
		return -1;
	
	int num = 0;

	char buffer[4096];
	const char* testBuffer;

	char* pos = strchr(token, '(');
	if(pos)
	{
		memcpy(buffer, token, pos-token);
		buffer[pos-token] = 0;
		testBuffer = buffer;
		*args = pos+1;

		*pos = 0;

		pos = strrchr(*args, ')');
		if(pos)
			*pos = 0;
	}
	else
	{
		testBuffer = token;
		*args = token;
	}

	while(idlist[num][0] != 0 && num < 20)
	{
		if(stricmp(testBuffer, idlist[num]) == 0)
		{
			return num;
		}
		num++;
	}
	return -1;
}

// RESOLVE PIDLE
inline bool spec(HWND owner, OLECHAR* name, int code, OLECHAR* filename, LPITEMIDLIST &pidl, char* path)
{
	if(_wcsicmp(filename, name))
		return false;

	if(FAILED( SHGetSpecialFolderLocation(owner, code, &pidl) ))
	{
		error("Failed to find folder '%S'",filename);
		return false;
	}

	if(path)
	SHGetPathFromIDList(pidl, path);
	return true;
}

bool Resolve(HWND owner, LPSHELLFOLDER psfDesktop, OLECHAR* filename, LPITEMIDLIST &pidl, char* path)
{
#define SPECIAL(name, code) else if(spec(owner, name, code, filename, pidl, path))return true;

	if(!filename)
		return false;
		SPECIAL(L"DESKTOP", CSIDL_DESKTOP)
		SPECIAL(L"INTERNET", CSIDL_INTERNET)
		SPECIAL(L"PROGRAMS", CSIDL_PROGRAMS)
		SPECIAL(L"CONTROLS", CSIDL_CONTROLS)
		SPECIAL(L"PRINTERS", CSIDL_PRINTERS)
		SPECIAL(L"PERSONAL", CSIDL_PERSONAL)
		SPECIAL(L"FAVORITES", CSIDL_FAVORITES)
		SPECIAL(L"STARTUP", CSIDL_STARTUP)
		SPECIAL(L"RECENT", CSIDL_RECENT)
		SPECIAL(L"SENDTO", CSIDL_SENDTO)
		SPECIAL(L"BIN", CSIDL_BITBUCKET)
		SPECIAL(L"STARTMENU", CSIDL_STARTMENU)
		SPECIAL(L"DESKTOPDIRECTORY", CSIDL_DESKTOPDIRECTORY)
		SPECIAL(L"DRIVES", CSIDL_DRIVES)
		SPECIAL(L"NETWORK", CSIDL_NETWORK)
		SPECIAL(L"NETHOOD", CSIDL_NETHOOD)
		SPECIAL(L"FONTS", CSIDL_FONTS)
		SPECIAL(L"TEMPLATES", CSIDL_TEMPLATES)
		SPECIAL(L"COMMON_STARTMENU", CSIDL_COMMON_STARTMENU)
		SPECIAL(L"COMMON_PROGRAMS", CSIDL_COMMON_PROGRAMS)
		SPECIAL(L"COMMON_STARTUP", CSIDL_COMMON_STARTUP)
		SPECIAL(L"COMMON_DESKTOPDIRECTORY", CSIDL_COMMON_DESKTOPDIRECTORY)
		SPECIAL(L"APPDATA", CSIDL_APPDATA)
		SPECIAL(L"PRINTHOOD", CSIDL_PRINTHOOD)
		SPECIAL(L"ALTSTARTUP", CSIDL_ALTSTARTUP)
		SPECIAL(L"COMMON_ALTSTARTUP", CSIDL_COMMON_ALTSTARTUP)
		SPECIAL(L"COMMON_FAVORITES", CSIDL_COMMON_FAVORITES)
		SPECIAL(L"INTERNET_CACHE", CSIDL_INTERNET_CACHE)
		SPECIAL(L"COOKIES", CSIDL_COOKIES)
		SPECIAL(L"HISTORY", CSIDL_HISTORY)
		else
	{
		if(FAILED(psfDesktop->ParseDisplayName(NULL, NULL, filename, NULL, &pidl, NULL)))
		{
			error("Failed to find folder '%S'", filename);
			return false;
		}
		return true;
	}
	
	return false;
}

LPITEMIDLIST MeldItems(LPITEMIDLIST first, LPITEMIDLIST second)
{	
	LPSHITEMID dir = &first->mkid;
	LPSHITEMID item = &second->mkid;

	int dirSize = GetItemSize(first);
	int itemSize = GetItemSize(second);

	int totalSize = dirSize + itemSize;

	LPITEMIDLIST ret = (LPITEMIDLIST)gMalloc->Alloc(totalSize + sizeof(SHITEMID));
	char* data = (char*)ret;

	memcpy(data, dir, dirSize);
	memcpy(data + dirSize, item, itemSize);

	LPSHITEMID terminator = (LPSHITEMID)(data + dirSize + itemSize);
	
	terminator->abID[0] = 0;
	terminator->cb = 0;

	return ret;
}

UINT GetItemSize  (LPITEMIDLIST pidl)
{
	USHORT nLen=0;
	while (pidl->mkid.cb != 0 && nLen < 4096)
	{
		nLen += pidl->mkid.cb;
		pidl = GetNextItem (pidl);
	}
	return nLen;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
// FUNCTION: GetItemCount
//
// DESCRIPTION: Computes the number of item IDs in an item ID list.
//
// INPUT: pidl = Pointer to an item ID list.
//
// RETURNS: Number of item IDs in the list.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

UINT GetItemCount (LPITEMIDLIST pidl)
{
	UINT nCount;
	nCount = 0;
	while (pidl->mkid.cb != 0 && nCount < 2000)
	{
		pidl = GetNextItem (pidl);
		nCount++;
	}
	return nCount;
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
// FUNCTION: GetNextItem
//
// DESCRIPTION: Finds the next item in an item ID list.
//
// INPUT: pidl = Pointer to an item ID list.
//
// RETURNS: Pointer to the next item.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

LPITEMIDLIST GetNextItem (LPITEMIDLIST pidl)
{
	USHORT nLen;
	
	if ((nLen = pidl->mkid.cb) == 0)
		return NULL;
	
	return (LPITEMIDLIST) (((LPBYTE) pidl) + nLen);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
// FUNCTION: DuplicateItem
//
// DESCRIPTION: Makes a copy of the next item in an item ID list.
//
// INPUT: gMalloc = Pointer to an IMalloc interface.
// pidl = Pointer to an item ID list.
//
// RETURNS: Pointer to an ITEMIDLIST containing the copied item ID.
//
// NOTES: It is the caller's responsibility to free the memory
// allocated by this function when the item ID is no longer
// needed. Example:
//
// pidlItem = DuplicateItem (gMalloc, pidl);
// .
// .
// .
// gMalloc->Free(pidlItem);
//
// Failure to free the ITEMIDLIST will result in memory
// leaks.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

LPITEMIDLIST DuplicateItem (LPITEMIDLIST pidl)
{
	USHORT nLen;
	LPITEMIDLIST pidlNew;
	
	nLen = pidl->mkid.cb;
	if (nLen == 0)
		return NULL;
	
	pidlNew = (LPITEMIDLIST) gMalloc->Alloc (nLen + sizeof (USHORT));
	if (pidlNew == NULL)
		return NULL;
	
	CopyMemory (pidlNew, pidl, nLen);
	*((USHORT*) (((LPBYTE) pidlNew) + nLen)) = 0;
	
	return pidlNew;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
// FUNCTION: DoExplorerMenu
//
// DESCRIPTION: Given a path name to a file or folder object, displays
// the shell's context menu for that object and executes
// the menu command (if any) selected by the user.
//
// INPUT: pszPath = Pointer to an ANSI or Unicode string
// specifying the path to the object.
//
//
// RETURNS: TRUE if successful, FALSE if not.
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

BOOL PopupContextMenu(LPITEMIDLIST pidlMain)
{
	LPSHELLFOLDER psfFolder, psfNextFolder;
	LPITEMIDLIST pidlItem, pidlNextItem, *ppidl;
	LPCONTEXTMENU pContextMenu;
	CMINVOKECOMMANDINFO ici;
	// ULONG ulCount, ulAttr;
	// char tchPath[MAX_PATH];
	// WCHAR wchPath[MAX_PATH];
	UINT nCount, nCmd;
	BOOL bResult=FALSE;
	HMENU hMenu;
	
	// Get current cursor position & window
	POINT point;
	HWND hwnd;
	
	GetCursorPos(&point);
	hwnd = WindowFromPoint(point);
	//
	
	if (!SUCCEEDED (SHGetDesktopFolder (&psfFolder)))
	{
		return bResult;
	}
	
	
	if (nCount = GetItemCount (pidlMain))
	{ // nCount must be > 0
		//
		// Initialize psfFolder with a pointer to the IShellFolder
		// interface of the folder that contains the item whose context
		// menu we're after, and initialize pidlItem with a pointer to
		// the item's item ID. If nCount > 1, this requires us to walk
		// the list of item IDs stored in pidlMain and bind to each
		// subfolder referenced in the list.
		//
		pidlItem = pidlMain;
		
		while (--nCount)
		{
			//
			// Create a 1-item item ID list for the next item in pidlMain.
			//
			pidlNextItem = DuplicateItem (pidlItem);
			if (pidlNextItem == NULL)
			{
				gMalloc->Free (pidlMain);
				psfFolder->Release();
				return bResult;
			}
			
			//
			// Bind to the folder specified in the new item ID list.
			//
			if (!SUCCEEDED (psfFolder->BindToObject (pidlNextItem, NULL, IID_IShellFolder, (void**)&psfNextFolder)))
			{
				gMalloc->Free(pidlNextItem);
				gMalloc->Free (pidlMain);
				psfFolder->Release();
				return bResult;
			}
			
			//
			// Release the IShellFolder pointer to the parent folder
			// and set psfFolder equal to the IShellFolder pointer for
			// the current folder.
			//
			psfFolder->Release();
			psfFolder = psfNextFolder;
			
			//
			// Release the storage for the 1-item item ID list we created
			// just a moment ago and initialize pidlItem so that it points
			// to the next item in pidlMain.
			//
			gMalloc->Free(pidlNextItem);
			pidlItem = GetNextItem (pidlItem);
		}
		
		//
		// Get a pointer to the item's IContextMenu interface and call
		// IContextMenu::QueryContextMenu to initialize a context menu.
		//
		ppidl = &pidlItem;
		if (SUCCEEDED (psfFolder->GetUIObjectOf (hwnd, 1, (const ITEMIDLIST**)ppidl, (const GUID)IID_IContextMenu, NULL, (void**)&pContextMenu)))
		{
			
			hMenu = CreatePopupMenu ();
			if (SUCCEEDED (pContextMenu->QueryContextMenu (
				hMenu, 0, 1, 0x7FFF, CMF_NORMAL)))
			{
				
				//
				// Display the context menu.
				//
				nCmd = TrackPopupMenu (hMenu, TPM_LEFTALIGN |
					TPM_LEFTBUTTON | TPM_RIGHTBUTTON | TPM_RETURNCMD,
					point.x, point.y, 0, hwnd, NULL);
				
				//
				// If a command was selected from the menu, execute it.
				//
				if (nCmd)
				{
					ici.cbSize = sizeof (CMINVOKECOMMANDINFO);
					ici.fMask = SEE_MASK_FLAG_NO_UI;
					ici.hwnd = hwnd;
					ici.lpVerb = MAKEINTRESOURCE (nCmd - 1);
					ici.lpParameters = NULL;
					ici.lpDirectory = NULL;
					ici.nShow = SW_SHOWNORMAL;
					ici.dwHotKey = 0;
					ici.hIcon = NULL;
					
					if (SUCCEEDED (
						pContextMenu->InvokeCommand (&ici)))
						bResult = TRUE;
				}
			}
			DestroyMenu (hMenu);
			pContextMenu->Release();
		}
	}
	gMalloc->Free (pidlMain);
	
	//
	// Clean up and return.
	//
	psfFolder->Release();
	// gMalloc->Release();
	
	return bResult;
}

/*
// retrieve a string value from the registry
LPTSTR RegQueryStringValue(HKEY hKey, LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pszBuffer, int nBufferSize, LPTSTR pszDefaultValue)
{
	DWORD dwDataType;
	DWORD dwDataSize = (DWORD) nBufferSize;
	HKEY hSubKey = NULL;
	LONG lErrorCode;
	
	if(pszSubKey)
	{
		lErrorCode = RegOpenKeyEx(hKey, pszSubKey, 0, KEY_QUERY_VALUE, &hSubKey);
		
		if(lErrorCode != ERROR_SUCCESS)
		{
			lstrcpyn(pszBuffer, pszDefaultValue, nBufferSize);
			return pszBuffer;
		}
		
		hKey = hSubKey;
	}
	
	lErrorCode = RegQueryValueEx(hKey, pszValueName, NULL, &dwDataType,
		(LPBYTE) pszBuffer, &dwDataSize);
	
	if(hSubKey)
		RegCloseKey(hSubKey);
	
	if(dwDataType != REG_SZ || lErrorCode != ERROR_SUCCESS)
		lstrcpyn(pszBuffer, pszDefaultValue, nBufferSize);
	
	return pszBuffer;
}

void TileBlt(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, HDC hdcSrc, int xSrc, int ySrc, int cxSrc, int cySrc, DWORD rasterOp)
{
	for(int y = 0; y < cyDest; y += cySrc)
	{
		for(int x = 0; x < cxDest; x += cxSrc)
		{
			BitBlt(hdcDest,
				xDest + x,
				yDest + y,
				min(cxDest - x, cxSrc),
				min(cyDest - y, cySrc),
				hdcSrc,
				xSrc,
				ySrc,
				rasterOp);
		}
	}
}

HDC hdcDesktop = 0;
HBITMAP hbmDesktop = 0;

// enhanced version of PaintDesktop
void PaintDesktopEx(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, int xSrc, int ySrc, BOOL updateCache)
{
	if(hdcDesktop == 0 || updateCache)
	{
		if(hdcDesktop == 0 && hdcDest == 0)
			return;
		
		if(hbmDesktop != 0)
		{
			hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);
			DeleteObject(hbmDesktop);
			hbmDesktop = 0;
		}
		
		HDC hdcScreen = GetDC(0);
		
		if(hdcDesktop == 0)
		{
			hdcDesktop = CreateCompatibleDC(hdcScreen);
		}
		
		int cxDesktop = GetSystemMetrics(SM_CXSCREEN);
		int cyDesktop = GetSystemMetrics(SM_CYSCREEN);
		
		hbmDesktop = CreateCompatibleBitmap(hdcScreen, cxDesktop, cyDesktop);
		hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);
		
		RECT r;
		SetRect(&r, 0, 0, cxDesktop, cyDesktop);
		FillRect(hdcDesktop, &r, GetSysColorBrush(COLOR_DESKTOP));
		
		char wallpaper[MAX_PATH] = { 0 };
		
		RegQueryStringValue(HKEY_CURRENT_USER,
			"Control Panel\\Desktop", "Wallpaper", wallpaper,
			MAX_PATH, "");
		
		if(wallpaper[0])
		{
			int wallpaperStyle;
			int tileWallpaper;
			
			char buffer[32];
			
			RegQueryStringValue(HKEY_CURRENT_USER, "Control Panel\\Desktop", "WallpaperStyle", buffer, 32, "0");
			wallpaperStyle = atoi(buffer);
			
			RegQueryStringValue(HKEY_CURRENT_USER, "Control Panel\\Desktop", "TileWallpaper", buffer, 32, "0");
			tileWallpaper = atoi(buffer);
			
			HBITMAP hbmWallpaper = (HBITMAP) LoadImage(0, wallpaper,
				IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
			
			if(hbmWallpaper)
			{
				BITMAP bm;
				GetObject(hbmWallpaper, sizeof(BITMAP), &bm);
				
				int cxWallpaper = bm.bmWidth;
				int cyWallpaper = bm.bmHeight;
				
				HDC hdcWallpaper = CreateCompatibleDC(hdcScreen);
				hbmWallpaper = (HBITMAP) SelectObject(hdcWallpaper, hbmWallpaper);
				
				if(tileWallpaper)
				{
					TileBlt(hdcDesktop,
						0, 0,
						cxDesktop, cyDesktop,
						hdcWallpaper,
						0, 0,
						cxWallpaper, cyWallpaper,
						SRCCOPY);
				}
				else
				{
					if(wallpaperStyle == 2) // stretch
					{
						SetStretchBltMode(hdcDesktop, STRETCH_DELETESCANS);
						
						StretchBlt(hdcDesktop,
							0, 0,
							cxDesktop, cyDesktop,
							hdcWallpaper,
							0, 0,
							cxWallpaper, cyWallpaper,
							SRCCOPY);
					}
					else
					{
						BitBlt(hdcDesktop,
							(cxDesktop - cxWallpaper) / 2,
							(cyDesktop - cyWallpaper) / 2,
							cxWallpaper, cyWallpaper,
							hdcWallpaper,
							0, 0,
							SRCCOPY);
					}
				}
				
				hbmWallpaper = (HBITMAP) SelectObject(hdcWallpaper, hbmWallpaper);
				DeleteDC(hdcWallpaper);
				
				DeleteObject(hbmWallpaper);
			}
		}
		
		ReleaseDC(0, hdcScreen);
 }
 
 if(hdcDest)
 {
	 BitBlt(hdcDest, xDest, yDest, cxDest, cyDest,
		 hdcDesktop, xSrc, ySrc, SRCCOPY);
 }
}
*/